;; Generated package description from jabber.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "jabber" "0.10.5.0.20260414.153707" "XMPP/Jabber client" '((emacs "29.1") (fsm "0.2.0")) :commit "0e43f5c5f23beaecf44321bba99bd422911bb593" :authors '(("Magnus Henoch" . "mange@freemail.hu")) :maintainer '("Thanos Apollo" . "public@thanosapollo.org") :keywords '("comm") :url "https://git.thanosapollo.org/emacs-jabber")
